# This package accepts a Perl graph and produces a Matlab m-file for use
# with the LSGS circuit solver
package Graph2lsgs;

use strict;
use warnings;
use Graph;

#custom packages
use GateLib;
use VerilogGraphMisc;
use v2tauMisc;

#Constructor
sub new {
    
    my $class = shift(@_);
    my $graph = shift(@_);
    my $nets = shift(@_);
    my $model_file = shift(@_);
    my $name_root = shift(@_);

    # the gates_only is an optional parameter
    # when set, it will only order the gates, but won't write out the m-file
    my $gates_only = shift(@_);

    my $nets_in = $nets->{in};

    my $gates = "";
    if (not defined $gates_only) {
	$gates = new GateLib($model_file);
    }

    ##OPEN FILES HERE
    my $filename_gates = "${name_root}_gates";
    my $filename_mfile = "${name_root}_lsgs.m";
    open WRITE_GATES, "> $filename_gates" or die "Can't open $filename_gates";

    if (not defined $gates_only) {
	open WRITE_MFILE, "> $filename_mfile" or 
	    die "Can't open $filename_mfile";
    }

    # these arrays contain the gate names for the gates in the F-matrix
    my @x = ();
    my @y = ();

    # these arrays contain the gate indices (i,j) and values (k) in the 
    # F-matrix
    my @i = ();
    my @j = ();
    my @k = ();

    # these arrays are the remaining LSGS parameters
    my @dmin = ();
    my @sideloads = ();
    my @area = ();
    my @x0 = ();

    # contains the list of gates in order
    my @gates_ordered = ();
    my @vertices_ordered = ();

    # a hash that maps gate names to indices
    my %indices = ();
    my $counter = 1;

    #arrays of current level and intermediate(next) level
    my @current = ();
    my @next = ();

    #the following block of code traverses the graph in a BFS manner
    #in order to ensure that the F matrix will be upper-triangular
    #this section is currently a performance bottleneck and so it has
    #been intentionally separated from the rest of the code so that it
    #can potentially be replaced.
######### START of BFS code    
    # traverse the graph in a BFS manner
    # start with gates whose inputs are ONLY primary inputs  
    foreach my $gate ($graph->successors("inputs")) {
	if (scalar($graph->predecessors($gate)) == 1) {
	    push(@current, $gate);
	    $indices{$gate} = $counter;
	    $counter = $counter + 1;
	}
    }

    # repeat until all gates have been assigned an index
    while (scalar(@current) > 0) {
	@next = ();
	# all gates in @current have been given an index
	foreach my $gate (@current) {
	    my @successors = $graph->successors($gate);
	    foreach my $successor (@successors) {

		# skip primary outputs
		if ($successor ne "outputs") {
		    # $numpred keeps track of how many predecessor gates still 
		    # have to be indexed. A gate is only indexed once all 
		    # predecessors have been indexed
		    my $numpred = $graph->get_vertex_attribute($successor, 
							       "numpred") - 1;
		    $graph->set_vertex_attribute($successor, "numpred", 
						 $numpred);   

		    if ($numpred == 0) {
			push(@next,$successor);
			$indices{$successor} = $counter;
			$counter = $counter + 1;
		    }
		}
	    }

	    push(@vertices_ordered, $gate);
	}
	@current = @next;
    }
######### END of BFS code    

    # we now have an appropriate ordering of the gates
    # go through gates (in order) and write out the vectors
    foreach my $gate (@vertices_ordered) {
	my $gate_name = VerilogGraphMisc::vertex2gate($gate);
	my $gate_size = VerilogGraphMisc::vertex2size($gate);
	my $gate_inst = VerilogGraphMisc::vertex2inst($gate);
	
	my @successors = $graph->successors($gate);

	my $di = 0;
	my $gi = 0;
	my $ai = 0;
	my $load_output = 0;

	foreach my $successor (@successors) {    
	    if ($successor eq "outputs") {
		$load_output = 1;
	    }
	    else {
		#determine the LE of the next gate
		my $successor_name = 
		    VerilogGraphMisc::vertex2gate($successor);
		
		my $port =
		    $graph->get_edge_attribute($gate, 
					       $successor, 
					       "in_pin");	
		my $le = 0;
		
		if (not defined $gates_only) {
		    $le = $gates->find_le($successor_name, $port);
		}
		
		push(@x,$gate);
		push(@y,$successor);
		push(@k,$le);	
	    }
	}

	if (not defined $gates_only) {
	    $di = $gates->find_parasitic($gate_name);
	    $gi = $graph->get_vertex_attribute($gate, "wireload");
	    $ai = $gates->find_area_slope($gate_name);
	}
	
	push(@dmin, $di);
	push(@sideloads, $gi);
	push(@area, $ai);
	push(@x0, $gate_size);   
	push(@gates_ordered, $gate_inst);
    }

    # convert gate names to indices
    foreach my $gate (@x) {
	push(@i, $indices{$gate});
    }
    foreach my $gate (@y) {
	push(@j, $indices{$gate});
    }
    
    my $date = get_date();
    my $ngates = scalar(@dmin);
    
    # Write the m-file 
    if (not defined $gates_only) {
	print WRITE_MFILE "% This m-file was autogenerated by v2tau.pl\n";
	print WRITE_MFILE "% on $date\n\n";
	# write sparse F matrix
	print_vector_matlab(\@i, "fx", \*WRITE_MFILE);
	print_vector_matlab(\@j, "fy", \*WRITE_MFILE);
	print_vector_matlab(\@k, "fval", \*WRITE_MFILE);
	print WRITE_MFILE "F = sparse(fx, fy, fval, $ngates, $ngates);\n\n";
	# write other vectors
	print_vector_matlab(\@dmin, "dmin", \*WRITE_MFILE);
	print_vector_matlab(\@sideloads, "g", \*WRITE_MFILE);
	print_vector_matlab(\@area, "a", \*WRITE_MFILE);
	print_vector_matlab(\@x0, "x0", \*WRITE_MFILE);
	close(WRITE_MFILE);
    }

    print_vector(\@gates_ordered, "gates", \*WRITE_GATES, 1);

    close(WRITE_GATES);

    return 0;
}


sub get_date {

    my @months = qw(Jan Feb Mar Apr May Jun Jul Aug Sep Oct Nov Dec);
    my @weekDays = qw(Sun Mon Tue Wed Thu Fri Sat Sun);
    my ($second, $minute, $hour, $dayOfMonth, $month, $yearOffset, $dayOfWeek, $dayOfYear, $daylightSavings) = localtime();
    my $year = 1900 + $yearOffset;
    my $theTime = "$hour:$minute:$second, $weekDays[$dayOfWeek] $months[$month] $dayOfMonth, $year";
    return $theTime;
}

sub print_vector {

    my $ref = shift(@_);
    my $name = shift(@_);
    my $file = shift(@_);

    foreach my $elem (@{$ref}) {
	print $file "$elem\n";
    }

}


sub print_vector_matlab {

    my $ref = shift(@_);
    my $name = shift(@_);
    my $file = shift(@_);

    print $file "$name = [...\n";
    foreach my $elem (@{$ref}) {
	print $file "$elem;\n";
    }
    print $file "];\n\n";
}

1;
